/*
 * File hash.c - generate hash tables for Wine debugger symbols
 *
 * Copyright (C) 1993, Eric Youngdale.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <elf.h>
#include "platform.h"

#if defined(TWIN_HASDLFCN)
#include <dlfcn.h>
#else
#include "twindlfcn.h"
#endif

extern FILE *debug_outfp;

struct  name_hash
{
    struct name_hash * next;
    unsigned int * address;
    char * name;
};

#define NR_NAME_HASH 128

static struct name_hash * name_hash_table[NR_NAME_HASH] = {0,};

static  unsigned int name_hash(const char * name)
{
    unsigned int hash = 0;
    const char * p;

    p = name;

    while (*p) 
	hash = (hash << 15) + (hash << 3) + (hash >> 3) + *p++;

    return hash % NR_NAME_HASH;
}


void add_hash(char * name, unsigned int * address)
{
    struct name_hash  * new;
    int hash;

    new = (struct  name_hash *) malloc(sizeof(struct name_hash));
    new->address = address;
    new->name = (char *) strdup(name);
    new->next = NULL;
    hash = name_hash(name);

    /* Now insert into the hash table */
    new->next = name_hash_table[hash];
    name_hash_table[hash] = new;
}

unsigned int * find_hash(char * name)
{
    char buffer[256];
    struct name_hash  * nh;

    for(nh = name_hash_table[name_hash(name)]; nh; nh = nh->next)
	if(strcmp(nh->name, name) == 0) 
	    return nh->address;

    if(name[0] != '_')
    {
	buffer[0] = '_';
	strcpy(buffer+1, name);
	for(nh = name_hash_table[name_hash(buffer)]; nh; nh = nh->next)
	    if(strcmp(nh->name, buffer) == 0) 
		return nh->address;
    }

    return (unsigned int *) 0xffffffff;
}

static char *dladdrtosym(void *addr, char *name_buffer)
{
#if TWIN_DLADDR
	/* if you have a working dladdr() */
	/* glibc2.0 dladdr is broken      */
    	Dl_info dlip;

	if(dladdr(addr, &dlip)) {

		add_hash(dlip.dli_sname,dlip.dli_saddr);

		strcpy(name_buffer, dlip.dli_sname);
		return name_buffer;
	} 
#endif

    return 0;
}


static struct name_hash * find_nearest_elf_symbol(void * address)
{
    static char name[256];
    static struct name_hash elf_nearest;

    if (!dladdrtosym(address, name))
	return NULL;

    elf_nearest.next = NULL;
    elf_nearest.address = find_hash(name);
    elf_nearest.name = name;
    
    return &elf_nearest;
}

char * find_nearest_symbol(unsigned int * address)
{
    struct name_hash *nearest = NULL;
    struct name_hash start;
    struct name_hash *nh;
    static char name_buffer[256];
    int i;

    nearest = find_nearest_elf_symbol(address);

    if (!nearest)
    {
	nearest = &start;
	start.address = (unsigned int *) 0;
    
        for(i=0; i<NR_NAME_HASH; i++) 
        {
	    for(nh = name_hash_table[i]; nh; nh = nh->next)
	        if (nh->address <= address && nh->address > nearest->address)
		   nearest = nh;
        }
        
        if ((unsigned int) nearest->address == 0) 
	   return NULL;
    }

    if (strlen(nearest->name)) {
        unsigned int delta = 
	    	(unsigned int) address - (unsigned int) nearest->address;
	if(delta)
        	sprintf(name_buffer, "%s+0x%x", nearest->name, delta);
	else
        	sprintf(name_buffer, "%s", nearest->name);
    	return name_buffer;
    }

    return NULL;
}

void
read_symboltable(char * filename)
{
    FILE * symbolfile;
    unsigned int addr;
    int nargs;
    char type;
    char * cpnt;
    char buffer[256];
    char name[256];
    static int initialized;

    if(initialized)
	return;

    initialized++;

    symbolfile = fopen(filename, "r");
    if(!symbolfile) 
    {
/*	fprintf(debug_outfp,"Unable to open symbol table %s\n", filename); */
	return;
    }

    fprintf(debug_outfp,"Reading symbols from file %s\n", filename);

    while (1)
    {
	fgets(buffer, sizeof(buffer),  symbolfile);
	if (feof(symbolfile)) break;
		
	/* Strip any text after a # sign (i.e. comments) */
	cpnt = buffer;
	while(*cpnt)
	{
	    if(*cpnt == '#') {*cpnt = 0; break; };
	    cpnt++;
	}
		
	/* Quietly ignore any lines that have just whitespace */
	cpnt = buffer;
	while(*cpnt)
	{
	    if(*cpnt != ' ' && *cpnt != '\t') 
		break;
	    cpnt++;
	}

	if (!(*cpnt) || *cpnt == '\n') 
	{
	    continue;
	}
		
	nargs = sscanf(buffer, "%x %c %s", &addr, &type, name);
	add_hash(name, (unsigned int *) addr);
    }

    fclose(symbolfile);
}

void find_invoke_native_bounds(unsigned long *startp,
			       unsigned long *endp)
{
    extern unsigned int * invoke_native;
    extern unsigned int * invoke_code16;

    add_hash("invoke_native",invoke_native);
    *startp = (unsigned long) find_hash("invoke_native");

    add_hash("invoke_code16",invoke_code16);
    *endp   = (unsigned long) find_hash("invoke_code16") - 1;

}
