/*    
	WinConfig.c	1.23
    	Copyright 1997 Willows Software, Inc. 

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.


For more information about the Willows Twin Libraries.

	http://www.willows.com	

To send email to the maintainer of the Willows Twin Libraries.

	mailto:twin@willows.com 
 */

/*************************************************************************
 *
 * Manage configuration parameters through a single configuration file.
 * 
 * These environment variables control parameters:
 *
 * TWINRC		location of configuration file if set
 * 			default, current working directory, home, willows
 * HOME			location of users home directory if set
 * 			default, value retrieved from pwnam
 * LOGNAME		user name if set
 *			default, use value from pwnam
 * DISPLAY		display server if set
 * 			default, value retrieved from twinrc file
 * PATH			path used to find and open files,
 *			default if not in TWINRC is to use PATH
 * LOGFILE		file to use for outputdebug string if set
 *			
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Driver.h"
#include "WinConfig.h"
#include "platform.h"

/**************************************************************************
 *
 * [boot]		startup configuration settings
 * [windows]		run-time configuration settings
 * [xdos]		mapping of DOS drives
 * [commondialogs]	how to drive common dialog fileopen dialog
 *
 * [fontalias]		what font mapping to use
 * [colors]		what colors to use
 *  
 **************************************************************************
 *
 *	parameter	what parameter is being addressed
 *	opcode	
 *		1	use environment as override if set
 *			use section, keyname, default to lookup string
 *		6	use environment as override if set
 *			use section, keyname, default to lookup string
 *			if null, return path to TWINRC
 *
 *	lpszsection	section to use
 *	lpszkeyname	keyname to use
 *	lpszdefault	default value
 *	lpszenviron	environment string
 *	lpszstring	allocated buffer
 *
 *************************************************************************/


static char *sections[] = {
	"boot",
	"windows",
	"XDOS",
	"CommonDialogs"
};

static TWINRC  *wcpdata = NULL;


/* name of configuration file */
extern BOOL TWIN_InitDriver();

char *
GetTwinFilename(void)
{
	static char *twinrc;

	/* have we already set it */
	if(twinrc)
		return twinrc;

	TWIN_InitDriver();
	twinrc = (char *) DRVCALL_CONFIG(PCFGH_GETFILENAME, 0, 0 , NULL);

	return twinrc;
	
}


char *
GetTwinBuffer(TWINRC *tp,char *wcpbuf)
{
	/* if we have a buffer, use it */
	if(wcpbuf)
		return wcpbuf;

	/* if we already allocated a buffer, return it */
	if(tp->lpszstring)
		return tp->lpszstring;

	/* allocate a permanent buffer */
	tp->lpszstring = (char *) WinMalloc(256);

	/* return it */
	return tp->lpszstring;	
}

static int Compatibility = -1;

char *
GetTwinString(int wcp, char *wcpbuf,int nsize )
{
	TWINRC  *tp;
	char     *p = NULL;
	static char *twinrc;

	if(twinrc == 0)
		twinrc = GetTwinFilename();
	
	if (wcpdata == NULL)
		wcpdata = (LPTWINRC)DRVCALL_CONFIG(PCFGH_GETDEFAULTS, 0, 0 , NULL);
	for(tp = wcpdata;tp->parameter > 0;tp++) {
		if(tp->parameter == wcp) {
		  switch(tp->opcode) {
			case 1:
				if(tp->lpszenviron && 
				   (p = (char *) getenv(tp->lpszenviron))) {
				       strcpy(GetTwinBuffer(tp,wcpbuf),p);

				       /* we got the environment variable */
				       return GetTwinBuffer(tp,wcpbuf);
				} 
				p = GetTwinBuffer(tp,wcpbuf);
				if(nsize == 0)
					nsize = 256;
				GetPrivateProfileString(
					sections[tp->section],
					tp->lpszkeyname,
					tp->lpszdefault,
					p,
					nsize, 
					twinrc);

				return GetTwinBuffer(tp,wcpbuf);

			 case 6:
				if(tp->lpszenviron && 
				   (p = (char *) getenv(tp->lpszenviron))) {
				       strcpy(GetTwinBuffer(tp,wcpbuf),p);

				       /* we got the environment variable */
				       return GetTwinBuffer(tp,wcpbuf);
				}
			
				p = GetTwinBuffer(tp,wcpbuf);
				if(nsize == 0)
					nsize = 256;
				/* get the value from the file */
				GetPrivateProfileString(
		  			sections[tp->section],
					tp->lpszkeyname,
					tp->lpszdefault,
					p,
	    				nsize, 
					twinrc);

				/* if we got something return it */
				if(strlen(p)) {
					return p;
				}
					
				/* use dirname(TwinFilename) */
				strcpy(p, GetTwinFilename());

				/* worst case same as TwinFilename directory */
				return p;
				
			default:
				break;
		  }
		  return wcpbuf;
		}
	}

	return 0;
}

unsigned long 
GetTwinInt(int wcp)
{
	char wcpbuffer[256];
	char *wcptr;

	wcptr = GetTwinString(wcp,wcpbuffer,256);
	
	return strtol(wcptr,0,0);
}

LPARAM
GetCompatibilityFlags(UINT uID)
{
    if ( Compatibility < 0 ) {
	Compatibility = 0;

	Compatibility = GetTwinInt(WCP_CONTROL);
    }

    return (LPARAM)Compatibility;
}

/********************************************************************
*	GetEnv
*
*	Our version of getenv that reads the [Environment] section of the twinrc file to obtain variables.
********************************************************************/
char *GetEnv(const char *env)
{
	char *twinrc = GetTwinFilename();
	static char buf[256];
	int count;

	/* Try and retrieve the environment variable */
	count = GetPrivateProfileString(sections[WCS_ENV], env, NULL, buf, 256,
		twinrc);

	if (count == 0)
		/* The default was used, meaning nothing was found */
		return(NULL);
	else
		/* Return the buffer containing the value found in the twinrc
		 * [Environment] section.
		 */
		return(buf);

}
